@@ -5,18 +5,20 @@ import (
 	"os"
 	"time"
 
-	kubemigratorclient "github.com/kubernetes-sigs/kube-storage-version-migrator/pkg/clients/clientset"
-	migrationv1alpha1informer "github.com/kubernetes-sigs/kube-storage-version-migrator/pkg/clients/informer"
 	configv1 "github.com/openshift/api/config/v1"
 	operatorv1 "github.com/openshift/api/operator/v1"
 	configv1client "github.com/openshift/client-go/config/clientset/versioned"
 	configv1informers "github.com/openshift/client-go/config/informers/externalversions"
+	operatorcontrolplaneclient "github.com/openshift/client-go/operatorcontrolplane/clientset/versioned"
+	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/audit"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/boundsatokensignercontroller"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/certrotationcontroller"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/certrotationtimeupgradeablecontroller"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/configmetrics"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/configobservation/configobservercontroller"
+	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/connectivitycheckcontroller"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/featureupgradablecontroller"
+	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/nodekubeconfigcontroller"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/operatorclient"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/resourcesynccontroller"
 	"github.com/openshift/cluster-kube-apiserver-operator/pkg/operator/targetconfigcontroller"
@@ -28,6 +30,7 @@ import (
 	"github.com/openshift/library-go/pkg/operator/encryption/controllers"
 	"github.com/openshift/library-go/pkg/operator/encryption/controllers/migrators"
 	encryptiondeployer "github.com/openshift/library-go/pkg/operator/encryption/deployer"
+	"github.com/openshift/library-go/pkg/operator/eventwatch"
 	"github.com/openshift/library-go/pkg/operator/genericoperatorclient"
 	"github.com/openshift/library-go/pkg/operator/resource/resourceapply"
 	"github.com/openshift/library-go/pkg/operator/staleconditions"
@@ -36,10 +39,14 @@ import (
 	"github.com/openshift/library-go/pkg/operator/staticresourcecontroller"
 	"github.com/openshift/library-go/pkg/operator/status"
 	"github.com/openshift/library-go/pkg/operator/v1helpers"
+	apiextensionsclient "k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
+	apiextensionsinformers "k8s.io/apiextensions-apiserver/pkg/client/informers/externalversions"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	"k8s.io/client-go/kubernetes"
+	kubemigratorclient "sigs.k8s.io/kube-storage-version-migrator/pkg/clients/clientset"
+	migrationv1alpha1informer "sigs.k8s.io/kube-storage-version-migrator/pkg/clients/informer"
 )
 
 type encryptionProvider []schema.GroupResource
@@ -64,6 +71,14 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 	if err != nil {
 		return err
 	}
+	operatorcontrolplaneClient, err := operatorcontrolplaneclient.NewForConfig(controllerContext.KubeConfig)
+	if err != nil {
+		return err
+	}
+	apiextensionsClient, err := apiextensionsclient.NewForConfig(controllerContext.KubeConfig)
+	if err != nil {
+		return err
+	}
 	kubeInformersForNamespaces := v1helpers.NewKubeInformersForNamespaces(
 		kubeClient,
 		"",
@@ -72,6 +87,7 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 		operatorclient.TargetNamespace,
 		operatorclient.OperatorNamespace,
 		"openshift-etcd",
+		"openshift-apiserver",
 	)
 	configInformers := configv1informers.NewSharedInformerFactory(configClient, 10*time.Minute)
 	operatorClient, dynamicInformers, err := genericoperatorclient.NewStaticPodOperatorClient(controllerContext.KubeConfig, operatorv1.GroupVersion.WithResource("kubeapiservers"))
@@ -89,14 +105,23 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 		return err
 	}
 
+	auditPolicyPahGetter, err := audit.NewAuditPolicyPathGetter()
+	if err != nil {
+		return err
+	}
 	configObserver := configobservercontroller.NewConfigObserver(
 		operatorClient,
 		kubeInformersForNamespaces,
 		configInformers,
 		resourceSyncController,
+		auditPolicyPahGetter,
 		controllerContext.EventRecorder,
 	)
 
+	eventWatcher := eventwatch.New().
+		WithEventHandler(operatorclient.TargetNamespace, "LateConnections", terminationobserver.ProcessLateConnectionEvents).
+		ToController(kubeInformersForNamespaces.InformersFor(operatorclient.TargetNamespace), kubeClient.CoreV1(), controllerContext.EventRecorder)
+
 	staticResourceController := staticresourcecontroller.NewStaticResourceController(
 		"KubeAPIServerStaticResources",
 		v410_00_assets.Asset,
@@ -104,9 +129,20 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 			"v4.1.0/kube-apiserver/ns.yaml",
 			"v4.1.0/kube-apiserver/svc.yaml",
 			"v4.1.0/kube-apiserver/kubeconfig-cm.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-clusterrole.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-clusterrole-node-reader.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-clusterrole-crd-reader.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-clusterrolebinding-auth-delegator.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-clusterrolebinding-node-reader.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-clusterrolebinding-crd-reader.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-kubeconfig-cm.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-rolebinding-kube-system.yaml",
+			"v4.1.0/kube-apiserver/check-endpoints-rolebinding.yaml",
+			"v4.1.0/kube-apiserver/control-plane-node-kubeconfig-cm.yaml",
 			"v4.1.0/kube-apiserver/localhost-recovery-client-crb.yaml",
 			"v4.1.0/kube-apiserver/localhost-recovery-sa.yaml",
 			"v4.1.0/kube-apiserver/localhost-recovery-token.yaml",
+			"v4.1.0/kube-apiserver/audit-policies-cm.yaml",
 		},
 		(&resourceapply.ClientHolder{}).WithKubernetes(kubeClient),
 		operatorClient,
@@ -123,6 +159,26 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 		controllerContext.EventRecorder,
 	)
 
+	nodeKubeconfigController := nodekubeconfigcontroller.NewNodeKubeconfigController(
+		operatorClient,
+		kubeInformersForNamespaces,
+		kubeClient,
+		configInformers.Config().V1().Infrastructures(),
+		controllerContext.EventRecorder,
+	)
+
+	apiextensionsInformers := apiextensionsinformers.NewSharedInformerFactory(apiextensionsClient, 10*time.Minute)
+	connectivityCheckController := connectivitycheckcontroller.NewKubeAPIServerConnectivityCheckController(
+		kubeClient,
+		operatorClient,
+		apiextensionsClient,
+		kubeInformersForNamespaces,
+		operatorcontrolplaneClient,
+		configInformers,
+		apiextensionsInformers,
+		controllerContext.EventRecorder,
+	)
+
 	// don't change any versions until we sync
 	versionRecorder := status.NewVersionGetter()
 	clusterOperator, err := configClient.ConfigV1().ClusterOperators().Get(ctx, "kube-apiserver", metav1.GetOptions{})
@@ -158,6 +214,7 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 			{Resource: "namespaces", Name: operatorclient.TargetNamespace},
 			{Group: "admissionregistration.k8s.io", Resource: "mutatingwebhookconfigurations"},
 			{Group: "admissionregistration.k8s.io", Resource: "validatingwebhookconfigurations"},
+			{Group: "controlplane.operator.openshift.io", Resource: "podnetworkconnectivitychecks", Namespace: "openshift-kube-apiserver"},
 		},
 
 		configClient.ConfigV1(),
@@ -256,11 +313,13 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 	configInformers.Start(ctx.Done())
 	dynamicInformers.Start(ctx.Done())
 	migrationInformer.Start(ctx.Done())
+	apiextensionsInformers.Start(ctx.Done())
 
 	go staticPodControllers.Start(ctx)
 	go resourceSyncController.Run(ctx, 1)
 	go staticResourceController.Run(ctx, 1)
 	go targetConfigReconciler.Run(ctx, 1)
+	go nodeKubeconfigController.Run(ctx, 1)
 	go configObserver.Run(ctx, 1)
 	go clusterOperatorStatus.Run(ctx, 1)
 	go certRotationController.Run(ctx, 1)
@@ -268,8 +327,10 @@ func RunOperator(ctx context.Context, controllerContext *controllercmd.Controlle
 	go featureUpgradeableController.Run(ctx, 1)
 	go certRotationTimeUpgradeableController.Run(ctx, 1)
 	go terminationObserver.Run(ctx, 1)
+	go eventWatcher.Run(ctx, 1)
 	go boundSATokenSignerController.Run(ctx, 1)
 	go staleConditionsController.Run(ctx, 1)
+	go connectivityCheckController.Run(ctx, 1)
 
 	<-ctx.Done()
 	return nil
@@ -296,6 +357,8 @@ var RevisionConfigMaps = []revision.RevisionResource{
 	{Name: "kube-apiserver-server-ca", Optional: true},
 	{Name: "kubelet-serving-ca"},
 	{Name: "sa-token-signing-certs"},
+
+	{Name: "kube-apiserver-audit-policies"},
 }
 
 // RevisionSecrets is a list of secrets that are directly copied for the current values.  A different actor/controller modifies these.
@@ -309,6 +372,8 @@ var RevisionSecrets = []revision.RevisionResource{
 	// this needs to be revisioned as certsyncer's kubeconfig isn't wired to be live reloaded, nor will be autorecovery
 	{Name: "localhost-recovery-serving-certkey"},
 	{Name: "localhost-recovery-client-token"},
+
+	{Name: "webhook-authenticator", Optional: true},
 }
 
 var CertConfigMaps = []revision.RevisionResource{
@@ -317,6 +382,12 @@ var CertConfigMaps = []revision.RevisionResource{
 
 	// this is a copy of trusted-ca-bundle CM without the injection annotations
 	{Name: "trusted-ca-bundle", Optional: true},
+
+	// kubeconfig that is a system:master.  this ensures a stable location
+	{Name: "control-plane-node-kubeconfig"},
+
+	// kubeconfig for check-endpoints
+	{Name: "check-endpoints-kubeconfig"},
 }
 
 var CertSecrets = []revision.RevisionResource{
@@ -326,6 +397,10 @@ var CertSecrets = []revision.RevisionResource{
 	{Name: "external-loadbalancer-serving-certkey"},
 	{Name: "internal-loadbalancer-serving-certkey"},
 	{Name: "bound-service-account-signing-key"},
+	{Name: "control-plane-node-admin-client-cert-key"},
+	{Name: "check-endpoints-client-cert-key"},
+
+	{Name: "node-kubeconfigs"},
 
 	{Name: "user-serving-cert", Optional: true},
 	{Name: "user-serving-cert-000", Optional: true},
